from __future__ import annotations

import hashlib
from enum import Enum
from types import MappingProxyType
from typing import List


class NodeType(Enum):
    ATTACK = 1
    DEFENCE = 2
    AND_GATE = 3


class Node(object):
    def __init__(self,
                 node_type: NodeType = NodeType.ATTACK,
                 label: str = "",
                 reference_id: str = "",
                 child_nodes: List[Node] = None,
                 metadata: dict = MappingProxyType({})):
        self.node_type = node_type
        self.label = label
        self.reference_id = reference_id
        self.child_nodes = [] if child_nodes is None else child_nodes
        self.metadata = {} | metadata

    def get_id(self) -> str:
        return hashlib.md5(self.label.encode()).hexdigest()

    def get_reference_id(self) -> str:
        return self.reference_id

    def get_node_type(self) -> NodeType:
        return self.node_type

    def get_label(self) -> str:
        return self.label

    def get_child_nodes(self) -> List[Node]:
        return self.child_nodes

    def add_metadata(self, key, value):
        self.metadata[key] = value

    def get_metadata(self, key):
        return self.metadata[key] if key in self.metadata else None

    def has_metadata(self, key):
        return key in self.metadata

    def __repr__(self):
        return f"{self.__class__.__name__}:{self.label}"


class Attack(Node):
    def __init__(self, label: str = "", child_nodes: List[Node] = None):
        super().__init__(node_type=NodeType.ATTACK,
                         label=label,
                         child_nodes=child_nodes)


class Defence(Node):
    def __init__(self, label: str = "", child_nodes: List[Node] = None):
        super().__init__(node_type=NodeType.DEFENCE,
                         label=label,
                         child_nodes=child_nodes)


class AndGate(Node):
    def __init__(self, child_nodes: List[Node] = None):
        super().__init__(node_type=NodeType.AND_GATE,
                         label="AND",
                         child_nodes=child_nodes)

    def get_id(self) -> str:
        # Set ID as the hash of all children labels
        children_labels = ""
        for child_node in self.child_nodes:
            children_labels += child_node.get_label()
        return hashlib.md5(children_labels.encode()).hexdigest()


class ADTree(Node):
    def __init__(self, reference_id: str = "", root_node: Node = None):
        super().__init__(node_type=root_node.node_type,
                         label=root_node.label,
                         reference_id=reference_id,
                         child_nodes=root_node.child_nodes)


class ExternalADTree(Node):
    def __init__(self, reference_id: str = "", label: str = ""):
        super().__init__(node_type=NodeType.ATTACK,
                         label=label,
                         reference_id=reference_id)



from __future__ import annotations

from graphviz import Graph

from adtree.models import Node, ADTree
from adtree.themes import Theme, NoFormatTheme


class Renderer(object):
    def __init__(self, theme: Theme = None, output_format: str = "png", view=False):
        self.theme = NoFormatTheme() if theme is None else theme
        self.output_format = output_format
        self.view = view
        self.references_to_expand = set()

    def render(self, tree: ADTree, subtrees_to_expand: list[ADTree] = None, filename: str = "attacktree-graph"):
        dot = Graph(graph_attr=self.theme.get_graph_attrs(tree),
                    format=self.output_format)
        dot.graph_attr["label"] = tree.get_reference_id() + " - " + tree.get_label()

        node_cache = set()
        references_to_expand = set()
        references_to_expand.add(tree.get_reference_id())
        if subtrees_to_expand is not None:
            for subtree in subtrees_to_expand:
                references_to_expand.add(subtree.get_reference_id())

        self._add_node(dot, tree, node_cache, references_to_expand)

        dot.render(filename, view=self.view)

    def _add_node(self, dot: Graph, current_node: Node, node_cache: set[str], references_to_expand: set[str]):
        node_attrs = self.theme.get_node_attrs_for(current_node)

        should_expand_node = True
        label = current_node.get_label()

        # Hide subtrees, unless explicitly told to expand
        if current_node.get_reference_id() != "" and \
                current_node.get_reference_id() not in references_to_expand:
            should_expand_node = False
            label += "\n(" + current_node.get_reference_id() + ")"

        dot.node(current_node.get_id(), label, **node_attrs)
        node_cache.add(current_node.get_id())

        if should_expand_node:
            for child_index, child_node in enumerate(current_node.get_child_nodes()):
                if child_node.get_id() not in node_cache:
                    self._add_node(dot, child_node, node_cache, references_to_expand)
                edge_attrs = self.theme.get_edge_attrs_for(current_node, child_node)
                dot.edge(current_node.get_id(), child_node.get_id(), **edge_attrs)



from __future__ import annotations

from abc import ABC, abstractmethod

from adtree.models import ADTree, Node, NodeType


class Analyser(ABC):
    @abstractmethod
    def analyse_tree(self, tree: ADTree) -> ADTree:
        pass


# Traverse the tree and mark each nodes as either defended or undefended
# A node is considered defended if:
# 1. is a Defence node and has no children
# 2. is an Attack node and has a direct defended Defence node as child
# 3. is an Attack or Defence node and all child nodes are defended nodes
# 4. is an AndGate and at least one child node is defended
class IsDefendedAnalyser(Analyser):
    METADATA_KEY = "IS_DEFENDED"

    def analyse_tree(self, tree: ADTree):
        self._process_node(tree)

    def _process_node(self, node: Node):
        # Process children first (depth first traversal)
        for child_node in node.get_child_nodes():
            self._process_node(child_node)

        # Precompute useful counters
        total_children = len(node.get_child_nodes())
        total_defended_nodes = len(list(filter(IsDefendedAnalyser._is_defended_node, node.get_child_nodes())))

        # Check all scenarios
        is_defended = False

        # 1. is a Defence node and has no children
        if node.get_node_type() == NodeType.DEFENCE:
            if total_children == 0:
                is_defended = True

        # 2. is an Attack node and has a direct defended Defence node as child
        if node.get_node_type() == NodeType.ATTACK:
            defended_nodes = filter(IsDefendedAnalyser._is_defended_node, node.get_child_nodes())
            defended_defence_nodes = filter(IsDefendedAnalyser._is_defence_node, defended_nodes)
            if len(list(defended_defence_nodes)) > 0:
                is_defended = True

        # 3. is an Attack or Defence node and all child nodes are defended nodes
        if node.get_node_type() == NodeType.ATTACK or node.get_node_type() == NodeType.ATTACK:
            if total_defended_nodes == total_children and total_children > 0:
                is_defended = True

        # 4. is an AndGate and at least one child node is defended
        if node.get_node_type() == NodeType.AND_GATE:
            if total_defended_nodes > 0:
                is_defended = True

        node.add_metadata(IsDefendedAnalyser.METADATA_KEY, is_defended)

    @staticmethod
    def _is_defended_node(node: Node):
        return node.get_metadata(IsDefendedAnalyser.METADATA_KEY)

    @staticmethod
    def _is_defence_node(node: Node):
        return node.get_node_type() == NodeType.DEFENCE



from __future__ import annotations

from abc import ABC, abstractmethod

from adtree.analysers import IsDefendedAnalyser
from adtree.models import Node, NodeType, ADTree


class Theme(ABC):
    @abstractmethod
    def get_graph_attrs(self, tree: ADTree) -> dict:
        pass

    @abstractmethod
    def get_node_attrs_for(self, node: Node) -> dict:
        pass

    @abstractmethod
    def get_edge_attrs_for(self, node_parent: Node, node_child: Node) -> dict:
        pass


class NoFormatTheme(Theme):

    def get_graph_attrs(self, tree: ADTree):
        return {}

    def get_node_attrs_for(self, node: Node):
        return {}

    def get_edge_attrs_for(self, node_parent: Node, node_child: Node):
        return {}


class BaseTheme(Theme):

    def get_graph_attrs(self, tree: ADTree):
        return {
            "overlap": "False",
            "splines": "True",
            "nodesep": "0.2",
            "ranksep": "0.4",
            "labelloc": "t",
            "labeljust": "l",
        }

    def get_node_attrs_for(self, node: Node):
        return {
            "color": "#000000",
            "fillcolor": "#ffffff",
            "shape": "box",
            "style": "rounded",
            "fontname": "Arial",
            "margin": "0.2"
        }

    def get_edge_attrs_for(self, node_parent: Node, node_child: Node):
        base_attrs = {
            "fontname": "Arial",
            "color": "#1f1f1f",
            "style": "solid"
        }

        if (node_parent.get_node_type() == NodeType.ATTACK and node_child.get_node_type() == NodeType.DEFENCE) or \
                (node_parent.get_node_type() == NodeType.DEFENCE and node_child.get_node_type() == NodeType.ATTACK):
            return base_attrs | {
                "style": "dashed"
            }
        else:
            return base_attrs


class RedBlueFillTheme(BaseTheme):

    def get_node_attrs_for(self, node: Node):
        base_attrs = super().get_node_attrs_for(node) | {
            "shape": "plaintext",
            "style": "filled, rounded"
        }

        if node.get_node_type() == NodeType.ATTACK:
            return base_attrs | {
                "fillcolor": "#ff5c5c",
            }
        elif node.get_node_type() == NodeType.DEFENCE:
            return base_attrs | {
                "fillcolor": "#5cc1ff",
            }
        elif node.get_node_type() == NodeType.AND_GATE:
            return base_attrs | {
                "shape": "triangle",
                "color": "#ff5c5c",
                "fillcolor": "#ff5c5c",
                "margin": "0.05"
            }
        else:
            return base_attrs


class RedGreenOutlineTheme(BaseTheme):

    def get_node_attrs_for(self, node: Node):
        default_style = "filled, setlinewidth(3)"

        base_attrs = super().get_node_attrs_for(node) | {
            "color": "#000000",
            "fillcolor": "#ffffff",
            "margin": "0.1",
            "style": default_style,
        }

        # Process attr based on node types
        node_type_attrs = base_attrs
        if node.get_node_type() == NodeType.ATTACK:
            node_type_attrs |= {
                "shape": "ellipse",
                "color": "#ff5c5c",
                "style": default_style,
            }
        elif node.get_node_type() == NodeType.DEFENCE:
            node_type_attrs |= node_type_attrs | {
                "shape": "box",
                "color": "#27B011",
                "style": default_style,
            }
        elif node.get_node_type() == NodeType.AND_GATE:
            node_type_attrs |= node_type_attrs | {
                "shape": "triangle",
                "color": "#ff5c5c",
                "margin": "0.05",
                "style": "rounded, "+default_style,
            }

        # Process attr based on metadata
        metadata_attrs = node_type_attrs
        if node.has_metadata(IsDefendedAnalyser.METADATA_KEY):
            fillcolor = "#C8FFCB" if node.get_metadata(IsDefendedAnalyser.METADATA_KEY) else "#FFD3D6"
            metadata_attrs |= {
                "fillcolor": fillcolor,
            }
        return metadata_attrs



from setuptools import setup

# !!!!!!! MAJOR DEBT - This is hardcoded
VERSION = "0.0.10"

# ~~~~~ Create configuration
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='adtree-viz',
    packages=[
        'adtree',
    ],
    package_dir={'': 'src'},
    install_requires=['graphviz==0.16'],
    version=VERSION,
    description='adtree-viz',
    long_description=(this_directory / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author='Julian Ghionoiu',
    author_email='julian.ghionoiu@gmail.com',
    url='https://github.com/julianghionoiu/adtree-viz',
    download_url='https://github.com/julianghionoiu/adtree-viz/archive/v{0}.tar.gz'.format(VERSION),
    keywords=['attack-defence', 'adtree'],
    classifiers=[],
)
